from .base import Base
import logging
from datetime import datetime
from fyle_accounting_mappings.models import ExpenseAttributesDeletionCache

logger = logging.getLogger(__name__)
logger.level = logging.INFO

class Categories(Base):
    """Class for Categories APIs."""

    def __init__(self):
        Base.__init__(self, attribute_type='CATEGORY')

    def sync(self, sync_after: datetime = None):
        """
        Syncs the latest API data to DB.
        """
        try:
            expense_attributes_deletion_cache, _ = ExpenseAttributesDeletionCache.objects.get_or_create(workspace_id=self.workspace_id)
            generator = self.get_all_generator()
            
            for items in generator:
                category_attributes = []
                expense_attributes_deletion_cache = ExpenseAttributesDeletionCache.objects.get(workspace_id=self.workspace_id)
                for category in items['data']:
                    expense_attributes_deletion_cache.category_ids.append(category['id'])
                    if self.attribute_is_valid(category):
                        if category['sub_category'] and category['name'] != category['sub_category']:
                            category['name'] = '{0} / {1}'.format(category['name'], category['sub_category'])

                        category_attributes.append({
                            'attribute_type': self.attribute_type,
                            'display_name': self.attribute_type.replace('_', ' ').title(),
                            'value': category['name'],
                            'source_id': category['id'],
                            'active': category['is_enabled'],
                            'detail': None
                        })

                expense_attributes_deletion_cache.save()
                self.bulk_create_or_update_expense_attributes(category_attributes, True)

            self.bulk_update_deleted_expense_attributes()

        except Exception as exception:
            logger.exception(exception)
            expense_attributes_deletion_cache = ExpenseAttributesDeletionCache.objects.get(workspace_id=self.workspace_id)
            expense_attributes_deletion_cache.category_ids = []
            expense_attributes_deletion_cache.save()



from .base import Base


class Employees(Base):
    """Class for Employees APIs."""

    def __init__(self):
        Base.__init__(self, attribute_type='EMPLOYEE', query_params={'is_enabled': 'eq.true'})

    def get_employee_by_email(self, email: str):
        """
        Get employee by email
        """
        return self.connection.list({
            'user->email': 'eq.{}'.format(email),
            'offset': 0,
            'limit': 1,
            'order': 'updated_at.desc'
        })['data']

    def get_admins(self):
        """
        Get admins of the org
        """
        admins = []
        generator = self.connection.list_all({
            'roles': 'like.%ADMIN%',
            'order': 'updated_at.desc'
        })

        for items in generator:
            for employee in items['data']:
                admins.append({
                    'user_id': employee['user']['id'],
                    'email': employee['user']['email'],
                    'full_name': employee['user']['full_name'],
                })

        return admins

    def sync(self):
        """
        Syncs the latest API data to DB.
        """
        generator = self.get_all_generator()
        for items in generator:
            employee_attributes = []
            for employee in items['data']:
                if self.attribute_is_valid(employee):
                    employee_attributes.append({
                        'attribute_type': self.attribute_type,
                        'display_name': self.attribute_type.replace('_', ' ').title(),
                        'value': employee['user']['email'],
                        'source_id': employee['id'],
                        'active': True,
                        'detail': {
                            'user_id': employee['user_id'],
                            'employee_code': employee['code'],
                            'full_name': employee['user']['full_name'],
                            'location': employee['location'],
                            'department': employee['department']['name'] if employee['department'] else None,
                            'department_id': employee['department_id'],
                            'department_code': employee['department']['code'] if employee['department'] else None
                        }
                    })

            self.bulk_create_or_update_expense_attributes(employee_attributes, True)



from .base import Base


class CostCenters(Base):
    """Class for Cost Centers APIs."""

    def __init__(self):
        Base.__init__(self, attribute_type='COST_CENTER', query_params={'is_enabled': 'eq.true'})



from datetime import datetime
from .base import Base

class TaxGroups(Base):
    """
    Class for Tax Groups API
    """

    def __init__(self):
        Base.__init__(self, attribute_type='TAX_GROUP', query_params={'order': 'id.asc'})

    def sync(self, sync_after: datetime = None):
        """
        Syncs the latest API data to DB.
        """
        generator = self.get_all_generator(sync_after)
        for items in generator:
            tax_attributes = []
            for tax_group in items['data']:
                    if self.attribute_is_valid(tax_group):
                        tax_attributes.append({
                            'attribute_type': 'TAX_GROUP',
                            'display_name': 'Tax Group',
                            'value': tax_group['name'],
                            'source_id': tax_group['id'],
                            'active': tax_group['is_enabled'],
                            'detail': {
                                'tax_rate': tax_group['percentage']
                            }
                        })

            self.bulk_create_or_update_expense_attributes(tax_attributes, True)



from .base import Base

class Reports(Base):

    def bulk_mark_as_paid(self, data):

        payload = {
            'data': data
        }

        return self.connection.bulk_mark_as_paid(payload)



from .base import Base


class DependentFields(Base):
    """
    Class For Expense Fields
    """

    def __init__(self):
        Base.__init__(self, attribute_type='EXPENSE_FIELDS')
    
    def set_connection(self, connection, expense_fields_connection):
        """
        Set connection
        """
        self.connection = connection
        self.expense_fields_connection = expense_fields_connection

    def get_project_field_id(self):
        """
        Get Project Field ID
        """
        query_params = {
            'limit': 1,
            'order': 'updated_at.desc',
            'offset': 0,
            'field_name': 'eq.Project',
            'is_custom': 'eq.False'
        }

        projects = self.expense_fields_connection.list(query_params)

        project_field_id = None

        if (len(projects['data'])) > 0:
            project_field_id = projects['data'][0]['id']

        return project_field_id


    def bulk_post_dependent_expense_field_values(self, data):
        """
        Post of Expense Field Values
        """
        payload = {
            'data': data
        }
        return self.connection.bulk_post_dependent_expense_field_values(payload)


    def sync(self):
        """
        Syncs the latest API data to DB.
        """
        generator = self.expense_fields_connection.list_all(
            query_params={
                'order': 'updated_at.desc',
                'is_enabled': 'eq.true',
                'is_custom': 'eq.true',
                'type': 'eq.DEPENDENT_SELECT'
            }
        )

        for items in generator:
            for row in items['data']:
                parent_field_id = row['parent_field_id']
                id = row['id']

                options_generator = self.connection.list_all(
                    {
                        'expense_field_id': f'eq.{id}',
                        'parent_expense_field_id': f'eq.{parent_field_id}',
                        'order': 'id.desc'
                    }
                )

                attributes = []
                count = 1
                attribute_type = row['field_name'].upper().replace(' ', '_')

                options_list = []
                for options in options_generator:
                    for option in options['data']:
                        if option['expense_field_value'] not in options_list:
                            options_list.append(option['expense_field_value'])

                            attributes.append({
                                'attribute_type': attribute_type,
                                'display_name': row['field_name'],
                                'value': option['expense_field_value'],
                                'active': True,
                                'source_id': 'expense_custom_field.{}.{}'.format(row['field_name'].lower(), count),
                                'detail': {
                                    'custom_field_id': row['id'],
                                    'placeholder': row['placeholder'],
                                    'is_mandatory': row['is_mandatory'],
                                    'is_dependent': True
                                }
                            })
                            count = count + 1

                self.attribute_type = attribute_type
                self.bulk_create_or_update_expense_attributes(attributes, True)


from .base import Base
from typing import List
from fyle_accounting_mappings.models import ExpenseAttribute


class Merchants(Base):
    """
    Class for Merchants API
    """
    def __init__(self):
        Base.__init__(self, attribute_type='MERCHANT', query_params={'column_name':'eq.merchant'})

    def get(self):
        generator = self.get_all_generator()
        for items in generator:
            merchants = items['data'][0]

        return merchants

    def post(self, payload: List[str], skip_existing_merchants: bool = False, delete_merchants: bool = False):
        """
        Post data to Fyle
        """
        generator = self.get_all_generator()
        for items in generator:
            merchants = items['data'][0]
            if delete_merchants:
                merchants['options'] = list(set(merchants['options']) - set(payload))
            else:
                if skip_existing_merchants:
                    merchants['options'] = payload
                else:
                    merchants['options'].extend(payload)
                merchants['options'] = list(set(merchants['options']))
            merchant_payload = {
                'id': merchants['id'],
                'field_name': merchants['field_name'],
                'type': 'SELECT',
                'options': merchants['options'],
                'placeholder': merchants['placeholder'],
                'category_ids': merchants['category_ids'],
                'is_enabled': merchants['is_enabled'],
                'is_custom': merchants['is_custom'],
                'is_mandatory': merchants['is_mandatory'],
                'code': merchants['code'],
                'default_value': merchants['default_value'] if merchants['default_value'] else '',
            }

        return self.connection.post({'data': merchant_payload})

    def sync(self):
        """
        Syncs the latest API data to DB.
        """
        generator = self.get_all_generator()
        for items in generator:
            merchants = items['data'][0]
            existing_merchants = ExpenseAttribute.objects.filter(
                attribute_type='MERCHANT', workspace_id=self.workspace_id)
            delete_merchant_ids = []

            if (existing_merchants):
                for existing_merchant in existing_merchants:
                    if existing_merchant.value not in merchants['options']:
                        delete_merchant_ids.append(existing_merchant.id)

                ExpenseAttribute.objects.filter(id__in=delete_merchant_ids).delete()

            merchant_attributes = []

            for option in merchants['options']:
                merchant_attributes.append({
                    'attribute_type': 'MERCHANT',
                    'display_name': 'Merchant',
                    'value': option,
                    'active': True,
                    'source_id': merchants['id'],
                })

            self.bulk_create_or_update_expense_attributes(merchant_attributes, True)



from .base import Base


class CorporateCards(Base):
    """Class for Corporate Cards APIs."""

    def __init__(self):
        Base.__init__(self, attribute_type='CORPORATE_CARD')

    def sync(self):
        """
        Syncs the latest API data to DB.
        """
        generator = self.get_all_generator()
        for items in generator:
            card_attributes = []
            unique_card_numbers = []
            for card in items['data']:
                if self.attribute_is_valid(card):
                    value = '{} - {}'.format(
                        card['bank_name'],
                        card['card_number'][-6:].replace('-', '')
                    )

                    if value not in unique_card_numbers:
                        unique_card_numbers.append(value)
                        card_attributes.append({
                            'attribute_type': self.attribute_type,
                            'display_name': self.attribute_type.replace('_', ' ').title(),
                            'value': value,
                            'source_id': card['id'],
                            'active': None,
                            'detail': {
                                'cardholder_name': card['cardholder_name']
                            }
                        })

            self.bulk_create_or_update_expense_attributes(card_attributes, True)



import logging
import os

from fyle.platform import Platform
from fyle.platform.exceptions import InvalidTokenError

from apps.workspaces.models import FyleCredential
from .apis import Expenses, Employees, Categories, Projects, CostCenters, ExpenseCustomFields, CorporateCards, \
    Reimbursements, TaxGroups, Merchants, Files, DependentFields, Departments, Subscriptions, Reports

logger = logging.getLogger(__name__)
logger.level = logging.INFO


class PlatformConnector:
    """The main class creates a connection with Fyle Platform APIs using OAuth2 authentication
    (refresh token grant type).

    Parameters:
    fyle_credential (str): Fyle Credential instance.
    """

    def __init__(self, fyle_credentials: FyleCredential):
        server_url = '{}/platform/v1beta'.format(fyle_credentials.cluster_domain)
        self.workspace_id = fyle_credentials.workspace_id
        try:
            self.connection = Platform(
                server_url=server_url,
                token_url=os.environ.get('FYLE_TOKEN_URI'),
                client_id=os.environ.get('FYLE_CLIENT_ID'),
                client_secret=os.environ.get('FYLE_CLIENT_SECRET'),
                refresh_token=fyle_credentials.refresh_token
            )
            
        except Exception:
            logger.info('Invalid refresh token')
            raise InvalidTokenError('Invalid refresh token')

        self.expenses = Expenses()
        self.employees = Employees()
        self.categories = Categories()
        self.projects = Projects()
        self.cost_centers = CostCenters()
        self.expense_custom_fields = ExpenseCustomFields()
        self.corporate_cards = CorporateCards()
        self.reimbursements = Reimbursements()
        self.tax_groups = TaxGroups()
        self.merchants = Merchants()
        self.files = Files()
        self.departments = Departments()
        self.dependent_fields = DependentFields()
        self.subscriptions = Subscriptions()
        self.reports = Reports()

        self.set_connection()
        self.set_workspace_id()

    def set_connection(self):
        """Set connection with Fyle Platform APIs."""
        self.expenses.set_connection(self.connection.v1beta.admin.expenses)
        self.employees.set_connection(self.connection.v1beta.admin.employees)
        self.categories.set_connection(self.connection.v1beta.admin.categories)
        self.projects.set_connection(self.connection.v1beta.admin.projects)
        self.cost_centers.set_connection(self.connection.v1beta.admin.cost_centers)
        self.expense_custom_fields.set_connection(self.connection.v1beta.admin.expense_fields)
        self.corporate_cards.set_connection(self.connection.v1beta.admin.corporate_cards)
        self.reimbursements.set_connection(self.connection.v1beta.admin.reimbursements)
        self.tax_groups.set_connection(self.connection.v1beta.admin.tax_groups)
        self.merchants.set_connection(self.connection.v1beta.admin.expense_fields)
        self.files.set_connection(self.connection.v1beta.admin.files)
        self.departments.set_connection(self.connection.v1beta.admin.departments)
        self.dependent_fields.set_connection(
            self.connection.v1beta.admin.dependent_expense_field_values,
            self.connection.v1beta.admin.expense_fields
        )
        self.subscriptions.set_connection(self.connection.v1beta.admin.subscriptions)
        self.reports.set_connection(self.connection.v1beta.admin.reports)

    def set_workspace_id(self):
        """Set workspace ID for Fyle Platform APIs."""
        self.expenses.set_workspace_id(self.workspace_id)
        self.employees.set_workspace_id(self.workspace_id)
        self.categories.set_workspace_id(self.workspace_id)
        self.projects.set_workspace_id(self.workspace_id)
        self.cost_centers.set_workspace_id(self.workspace_id)
        self.expense_custom_fields.set_workspace_id(self.workspace_id)
        self.corporate_cards.set_workspace_id(self.workspace_id)
        self.reimbursements.set_workspace_id(self.workspace_id)
        self.tax_groups.set_workspace_id(self.workspace_id)
        self.merchants.set_workspace_id(self.workspace_id)
        self.files.set_workspace_id(self.workspace_id)
        self.departments.set_workspace_id(self.workspace_id)
        self.dependent_fields.set_workspace_id(self.workspace_id)
        self.reports.set_workspace_id(self.workspace_id)

    def import_fyle_dimensions(self, import_taxes: bool = False, import_dependent_fields: bool = False, is_export: bool = False):
        """Import Fyle Platform dimension."""
        apis = ['employees', 'categories', 'projects', 'cost_centers', 'expense_custom_fields', 'corporate_cards']

        if is_export:
            apis = ['employees', 'cost_centers', 'expense_custom_fields', 'corporate_cards']

        if import_dependent_fields:
            apis.append('dependent_fields')

        if import_taxes:
            apis.append('tax_groups')

        for api in apis:
            dimension = getattr(self, api)
            try:
                dimension.sync()
            except Exception as e:
                logger.exception(e)
